""" Calculate the entropy and perplexity of a corpus."""

from typing import Callable, Dict, Union

import numpy as np
from spacy.language import Language
from spacy.lookups import load_lookups
from spacy.tokens import Doc, Span
from spacy.vocab import Vocab
from wasabi import msg


def set_lexeme_prob_table(vocab: Vocab, verbose: bool = False):
    """Ensure that the pipeline has a lexeme probability table."""
    if not vocab.lookups.has_table("lexeme_prob"):
        if verbose:
            msg.info(
                "Pipeline does not have a lexeme probability table. "
                + "Will attempt to add it. This will fail if specified language does "
                + "not have a registered lexeme probability table.",
            )
        lookups = load_lookups(vocab.lang, ["lexeme_prob"])
        vocab.lookups.add_table("lexeme_prob", lookups.get_table("lexeme_prob"))


def entropy(log_probs=np.ndarray) -> float:
    """Calculates the Shannon entropy based on log probs."""
    return -np.sum(np.exp(log_probs) * log_probs)


def perplexity(entropy: float) -> float:
    """Calculates the perplexity.

    Calculated as exp(H(p)), where H is the entropy using a base e and p is the
    probabilities of a given word.
    """
    return np.exp(entropy)


def entropy_getter(doc: Union[Doc, Span], log_prob_attr: str = "prob") -> float:
    """Calculate the shannon entropy of a document."""
    log_probs = np.array([getattr(token, log_prob_attr) for token in doc])
    return entropy(log_probs)


def perplexity_getter(doc: Union[Doc, Span]) -> float:
    """Calculates the perplexity of a doc.

    Calculated as exp(H(p)), where H is the entropy using a base e and p is the
    probabilities of a given word.
    """
    # check if it has the attribute entropy
    if hasattr(doc._, "entropy") and doc._.entropy is not None:
        entropy = doc._.entropy
    else:
        entropy = entropy_getter(doc)
    return perplexity(entropy)


def per_word_perplexity_getter(doc: Union[Doc, Span]) -> float:
    """Calculates the per word perplexity of a document."""
    if hasattr(doc._, "perplexity") and doc._.perplexity is not None:
        perplexity = doc._.perplexity
    else:
        perplexity = perplexity_getter(doc)

    return perplexity / len(doc)


def set_docspan_extension(
    extension,
    getter: Callable,
    prefix: str = "",
    force: bool = False,
) -> None:
    """Set a document extension to None and set a span extension to a getter
    function."""
    if not Doc.has_extension(prefix + extension) or force:
        Doc.set_extension(prefix + extension, default=None, force=True)
    if not Span.has_extension(prefix + extension) or force:
        Span.set_extension(prefix + extension, getter=getter)


def set_entropy_and_perplexity(doc: Union[Doc, Span]) -> None:
    doc._.entropy = entropy_getter(doc)
    doc._.perplexity = perplexity_getter(doc)
    doc._.per_word_perplexity = per_word_perplexity_getter(doc)


def set_entropy_and_perplexity_to_nan(doc: Union[Doc, Span]) -> None:
    doc._.entropy = np.nan
    doc._.perplexity = np.nan
    doc._.per_word_perplexity = np.nan


class InformationTheory:
    """SpaCy component for adding information theoretic metrics such as entropy
    and perplexity."""

    def __init__(self, nlp: Language, name: str, force: bool) -> None:
        self.name = name
        self.set_extensions(force=force)
        try:
            set_lexeme_prob_table(nlp.vocab, verbose=False)
            self.has_lexeme_prob_table = True
        except ValueError:
            msg.warn(
                f"Could not load lexeme probability table for language {nlp.lang}. "
                + "This will result in NaN values for perplexity and entropy.",
            )
            self.has_lexeme_prob_table = False

    @staticmethod
    def dict_getter(doc: Union[Doc, Span]) -> Dict[str, float]:
        return {
            "entropy": doc._.entropy,
            "perplexity": doc._.perplexity,
            "per_word_perplexity": doc._.per_word_perplexity,
        }

    @staticmethod
    def set_extensions(force: bool) -> None:
        """Add entropy and perplexity as attributes to a document."""
        for ext, getter in [
            ("entropy", entropy_getter),
            ("perplexity", perplexity_getter),
            ("per_word_perplexity", per_word_perplexity_getter),
            ("information_theory", InformationTheory.dict_getter),
        ]:
            set_docspan_extension(
                ext,
                getter=getter,  # type: ignore
                prefix="",
                force=force,
            )

    def __call__(self, doc: Doc) -> Doc:
        if self.has_lexeme_prob_table:
            set_entropy_and_perplexity(doc)
        else:
            set_entropy_and_perplexity_to_nan(doc)
        doc._.information_theory = InformationTheory.dict_getter(doc)
        return doc


@Language.factory(
    "textdescriptives/information_theory",
    assigns=[
        "doc._.entropy",
        "doc._.perplexity",
        "doc._.per_word_perplexity",
        "span._.entropy",
        "span._.perplexity",
        "span._.per_word_perplexity",
    ],
)
def create_information_theory_component(nlp: Language, name: str) -> InformationTheory:
    """
    Allows the InformationTheory component to be added to the spaCy pipeline using the
    command: `nlp.add_pipe('textdescriptives/information_theory')`

    It also set the following attributes on the document and span:

    - {Doc/Span}._.entropy: The shannon entropy of the document.
    - {Doc/Span}._.perplexity: The perplexity of the document.
    - {Doc/Span}._.per_word_perplexity: The per word perplexity of the document.
    - {Doc/Span}._.information_theory: A dictionary with the
        keys: entropy, perplexity, and per_word_perplexity.

    Args:
        - nlp: The spaCy Language object.
        - name: The name of the component.

    Example:
        >>> import spacy
        >>> import textdescriptives as td
        >>> nlp = spacy.blank('en')
        >>> nlp.add_pipe('textdescriptives/information_theory')
        >>> doc = nlp('This is a sentence.')
        >>> doc._.information_theory
        {'entropy': ...
    """
    return InformationTheory(nlp, name, force=False)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

from textdescriptives.about import __version__

# -- Project information -----------------------------------------------------

project = "textdescriptives"
author = "Lasse Hansen, Kenneth Enevoldsen"

# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxext.opengraph",
    "sphinx_copybutton",
    "sphinx.ext.githubpages",
    "sphinxcontrib.autodoc_pydantic",
    "myst_nb",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx_design",
]

# autodoc_mock_imports = ["textdescriptives"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# Autodoc Pydantic
autodoc_pydantic_model_show_json = False
autodoc_pydantic_model_show_config = False
autodoc_pydantic_model_show_field_summary = False


# -- Options for myst-nb -----------------------------------------------------

# set the timeout for executing notebooks
# nb_execution_timeout = 600  # in seconds, default 30 seconds
nb_execution_timeout = 600  # in seconds, default 30 seconds. 600 seconds = 10 minutes

# always fail CI pipeline when nb cannot be executed
nb_execution_raise_on_error = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_show_sourcelink = True

source_suffix = {".rst": "restructuredtext", ".md": "myst-nb", ".ipynb": "myst-nb"}

html_context = {
    "display_github": True,  # Add 'Edit on Github' link instead of 'View page source'
    "github_user": "HLasse",
    "github_repo": project,
    "github_version": "main",
    "conf_py_path": "/docs/",
}

html_theme_options = {
    "light_logo": "icon.png",
    "dark_logo": "icon_dark.png",
    "light_css_variables": {
        "color-brand-primary": "#ff5454",
        "color-brand-content": "#ff7575",
    },
    "dark_css_variables": {
        "color-brand-primary": "#ff8f8f",
        "color-brand-content": "#ff8f8f",
    },
    "sidebar_hide_name": True,
    "navigation_with_keys": True,
}

# pygments_style = "monokai"
# pygments_dark_style = "monokai"



""" Extract metrics as Pandas DataFrame."""

from typing import Any, Dict, Iterable, List, Optional, Union

import pandas as pd
from spacy.tokens import Doc
from wasabi import msg

from textdescriptives.utils import (
    _create_spacy_pipeline,
    _remove_textdescriptives_extensions,
    get_valid_metrics,
)


def __get_quality(doc: Doc) -> dict:
    """Get quality metrics as well as boolean indicator for passing filters."""
    return doc._.quality.to_flat_value_dict()


def __get_descriptive_stats_dict(doc: Doc) -> dict:
    """Get descriptive statistics as dictionary."""
    return {
        **doc._.token_length,
        **doc._.sentence_length,
        **doc._.syllables,
        **doc._.counts,
    }


def extract_dict(
    docs: Union[Iterable[Doc], Doc],
    metrics: Union[List[str], str, None] = None,
    include_text: bool = True,
) -> List[Dict[str, Any]]:
    """Extract calculated metrics from a spaCy Doc or an iterable of Docs to a
    list of dictionaries.

    Args:
        docs (Union[Iterable[Doc],  Doc]): An iterable of spaCy Docs or a single Doc
        metrics (Union[list[str], str, None], optional): Which metrics to extract.
            One or more of ["descriptive_stats", "readability",
            "dependency_distance", "pos_proportions", "coherence", "quality",
            "information_theory"]. Defaults to None in which case it will
            extract metrics for which a pipeline compoenent has been set.
        include_text (bool, optional): Whether to add an entry containing the text.
            Defaults to True.

    Returns:
        List[Dict[str, Any]]: List of dictionaries for each Doc with extracted metrics.
    """
    if not isinstance(docs, Doc):
        return [extract_dict(doc, metrics, include_text)[0] for doc in docs]

    # extract textdescriptive metrics from the list of spacy Language factory
    valid_metrics = get_valid_metrics()

    if isinstance(metrics, str):
        metrics = [metrics]

    if metrics is None:
        metrics = [
            component for component in valid_metrics if docs.has_extension(component)
        ]

    if not set(metrics).issubset(valid_metrics):
        raise ValueError(
            "'metrics' contained invalid metric.\n"
            + f"Valid metrics are: {valid_metrics}",
        )
    extracted_metrics: Dict[str, Any] = {}
    if include_text:
        extracted_metrics["text"] = docs.text
    for component in metrics:
        if component == "quality":
            metric = __get_quality(docs)
        elif component == "descriptive_stats":
            metric = __get_descriptive_stats_dict(docs)
        else:
            metric = getattr(docs._, component)
        if metric:
            extracted_metrics.update(metric)

    return [extracted_metrics]


def extract_df(
    docs: Union[Iterable[Doc], Doc],
    metrics: Union[List[str], str, None] = None,
    include_text: bool = True,
) -> pd.DataFrame:
    """Extract calculated metrics from a spaCy Doc object or a generator of Docs
    from nlp.pipe to a Pandas DataFrame.

    Args:
        docs (Union[Iterable[Doc],  Doc]): An iterable of spaCy Docs or a single Doc
        metrics (Union[list[str], str], optional): Which metrics to extract.
            One or more of ["descriptive_stats", "readability",
            "dependency_distance", "pos_proportions", "coherence", "quality",
            "information_theory"]. Defaults to None in which
            case it will extract metrics for which a pipeline compoenent has been
            set.
        include_text (bool, optional): Whether to add a column containing the text.
            Defaults to True.

    Returns:
        pd.DataFrame: DataFrame with a row for each doc and column for each metric.
    """
    return pd.DataFrame(extract_dict(docs, metrics, include_text))


def extract_metrics(
    text: Union[str, List[str]],
    lang: Optional[str] = None,
    metrics: Optional[Iterable[str]] = None,
    spacy_model: Optional[str] = None,
    spacy_model_size: str = "lg",
) -> pd.DataFrame:
    """Extract metrics from a text or a list of texts to a Pandas dataframe.

    Args:
        text (Union[str, List[str]]): A text or a list of texts.
        lang (str, optional): Language of the text. If lang is set and no spacy
            model is provided, will automatically download and use a spacy
            model for the language. Defaults to None.
        metrics (List[str]): Which metrics to extract.
            One or more of ["descriptive_stats", "readability",
            "dependency_distance", "pos_proportions", "coherence", "quality",
            "information_theory"]. If None, will extract all metrics from
            textdescriptives. Defaults to None.
        spacy_model (str, optional): The spacy model to use. If not set,
            will download one based on lang. Defaults to None.
        spacy_model_size (str, optional): Size of the spacy model to download.

    Returns:
        pd.DataFrame: DataFrame with a row for each text and column for each metric.
    """
    if isinstance(metrics, str):
        metrics = [metrics]

    if spacy_model is None and lang is None:
        raise ValueError("Either a spacy model or a language must be provided.")

    if spacy_model is not None and lang is not None:
        msg.info(
            "Both a spacy model and a language were provided. "
            + "Will use the spacy model and ignore language.",
        )

    if metrics is None:
        metrics = get_valid_metrics()

    # remove previously set metrics to avoid conflicts
    _remove_textdescriptives_extensions()

    # load spacy model if any component requires it
    nlp = _create_spacy_pipeline(
        spacy_model=spacy_model,
        lang=lang,
        metrics=metrics,
        spacy_model_size=spacy_model_size,
    )

    # add pipeline components
    if "all" in metrics:
        nlp.add_pipe("textdescriptives/all")
    else:
        for component in metrics:
            nlp.add_pipe(f"textdescriptives/{component}")

    if isinstance(text, str):
        text = [text]
    docs = nlp.pipe(text)

    return extract_df(docs)



import textdescriptives as td


def test_get_valid_metrics():
    metrics = td.get_valid_metrics()
    assert isinstance(metrics, set)
    assert len(metrics) > 0
    assert isinstance(metrics.pop(), str)


def test_get_columns():
    columns_ = set()
    for metric in td.get_valid_metrics():
        columns = td.get_doc_assigns(metric)
        assert isinstance(columns, list)
        assert len(columns) > 0
        assert isinstance(columns[0], str)
        columns_.update(columns)
    columns_all = td.get_doc_assigns("all")
    assert set(columns_all) == columns_



""" Calculation of statistics related to dependency distance."""

from typing import Callable

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc, Span, Token


class DependencyDistance:
    """spaCy v.3.0 component that adds attributes to `Doc`, `Span`, and `Token`
    objects relating to dependency distance.

    Dependency distance can be used as a measure of syntactic complexity, and
    measures the distance from a word to its head word. For `Doc` objects,
    dependency distance is calculated on the sentence level.
    """

    def __init__(self, nlp: Language):
        """Initialise components."""
        if not Token.has_extension("dependency_distance"):
            Token.set_extension("dependency_distance", getter=self.token_dependency)
        if not Span.has_extension("dependency_distance"):
            Span.set_extension("dependency_distance", getter=self.span_dependency)
        if not Doc.has_extension("dependency_distance"):
            Doc.set_extension("dependency_distance", getter=self.doc_dependency)

    def token_dependency(self, token: Token) -> dict:
        """Calculate token level dependency distance, i.e. the distance from a
        token to its head token. Also returns a boolean indicating whether the
        dependency relation is adjacent to the token.

        Returns:
            dict: Dictionary with the following keys:
                - dependency_distance: Dependency distance
                - adjacent_dependency: Boolean indicating whether the dependency
                  relation is adjacent to the token
        """
        dep_dist = 0
        ajd_dep = False
        if token.dep_ != "ROOT":
            dep_dist = abs(token.head.i - token.i)
            if dep_dist == 1:
                ajd_dep = True
        return {"dependency_distance": dep_dist, "adjacent_dependency": ajd_dep}

    def span_dependency(self, span: Span) -> dict:
        """Aggregates token level dependency distance on the span level by
        taking the mean of the dependency distance and the proportion of
        adjacent dependency relations.

        Returns:
            dict: Dictionary with the following keys: dependency_distance_mean:
                Mean dependency distance and prop_adjacent_dependency_relation:
                Proportion of adjacent dependency relations
        """
        dep_dists, adj_deps = zip(
            *[token._.dependency_distance.values() for token in span],
        )
        return {
            "dependency_distance_mean": np.mean(dep_dists),
            "prop_adjacent_dependency_relation": np.mean(adj_deps),
        }

    def doc_dependency(self, doc: Doc) -> dict:
        """Aggregates token level dependency distance on the document level by
        taking the mean of the dependency distance and the proportion of
        adjacent dependency relations on the sentence level.

        Returns:
            dict: Dictionary with the following keys:
                - dependency_distance_mean: Mean dependency distance on the sentence
                  level
                - dependency_distance_std: Standard deviation of dependency distance on
                  the sentence level
                - prop_adjacent_dependency_relation_mean: Mean proportion of adjacent
                  dependency relations on the sentence level
                - prop_adjacent_dependency_relation_std: Standard deviation of
                  proportion of adjacent dependency relations on the sentence level
        """
        if len(doc) == 0:
            return {
                "dependency_distance_mean": np.nan,
                "dependency_distance_std": np.nan,
                "prop_adjacent_dependency_relation_mean": np.nan,
                "prop_adjacent_dependency_relation_std": np.nan,
            }
        dep_dists, adj_deps = zip(
            *[sent._.dependency_distance.values() for sent in doc.sents],
        )
        return {
            "dependency_distance_mean": np.mean(dep_dists),
            "dependency_distance_std": np.std(dep_dists),
            "prop_adjacent_dependency_relation_mean": np.mean(adj_deps),
            "prop_adjacent_dependency_relation_std": np.std(adj_deps),
        }

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        return doc


@Language.factory(
    "textdescriptives/dependency_distance",
    assigns=[
        "token._.dependency_distance",
        "span._.dependency_distance",
        "doc._.dependency_distance",
    ],
)
def create_dependency_distance_component(
    nlp: Language,
    name: str,
) -> Callable[[Doc], Doc]:
    """Create spaCy language factory that allows DependencyDistance attributes
    to be added to a pipe using
    nlp.add_pipe("textdescriptives/dependency_distance")

    Adding this component to a pipeline sets the following attributes:
        - `token._.dependency_distance`
        - `span._.dependency_distance`
        - `doc._.dependency_distance`

    Args:
        nlp (Language): spaCy language object, does not need to be specified in the
            nlp.add_pipe call.
        name (str): name of the component. Can be optionally specified in the
            nlp.add_pipe call, using the name argument.

    Returns:
        Callable[[Doc], Doc]: The DependencyDistance component

    Example:
        >>> import spacy
        >>> import textdescriptives as td
        >>> nlp = spacy.load("en_core_web_sm")
        >>> nlp.add_pipe("textdescriptives/dependency_distance")
        >>> # apply the pipeline to a text
        >>> doc = nlp("This is a sentence.")
        >>> # access the dependency distance attributes
        >>> doc._.dependency_distance
    """
    return DependencyDistance(nlp)



import numpy as np
import pytest
import spacy

import textdescriptives as td  # noqa: F401
import warnings


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives/coherence")
    return nlp


def test_coherence_integration(nlp):
    assert "textdescriptives/coherence" == nlp.pipe_names[-1]


def test_coherence(nlp):
    doc = nlp(
        "This is a short and simple sentence. Here is yet another one. We need quite "
        + "a few before these coherences metrics make sense. Rambling, on and on.",
    )

    assert doc._.coherence
    assert doc._.first_order_coherence_values
    assert doc._.second_order_coherence_values


def test_coherence_single_sentence(nlp):
    doc = nlp("This is a short and simple sentence.")

    assert np.isnan(doc._.first_order_coherence_values).all()
    assert np.isnan(doc._.second_order_coherence_values).all()


def test_coherence_difference(nlp):
    coherent_doc = nlp(
        "We will now talk about animals. Dogs are animals. Cats are animals. Birds are"
        + " animals. Fish are animals.",
    )
    incoherent_doc = nlp(
        "Let's talk about a bunch of things. Houses made of pancakes and dogs talking"
        + " like humans. Look, the snow is falling.",
    )
    assert (
        coherent_doc._.coherence["first_order_coherence"]
        > incoherent_doc._.coherence["first_order_coherence"]
    )
    assert (
        coherent_doc._.coherence["second_order_coherence"]
        > incoherent_doc._.coherence["second_order_coherence"]
    )


def test_coherence_multi_process(nlp):
    docs = nlp.pipe(
        [
            "This is a short and simple sentence. Here is yet another one. We need "
            + "quite a few before these coherences metrics make sense. Rambling, on "
            + "and on.",
            "And another one. That's it. No more.",
        ],
        n_process=2,
    )
    for doc in docs:
        assert doc._.coherence


def test_coherence_blank_pipe():
    nlp = spacy.blank("en")
    nlp.add_pipe("textdescriptives/coherence")

    text = "This is a short and simple sentence. Here is yet another one."
    # check that an exception is raised
    with pytest.raises(ValueError) as e:
        doc = nlp(text)  # noqa F841
    assert "sentence boundary detector has not been" in str(e.value)

    # add a sentencizer
    nlp.add_pipe("sentencizer", before="textdescriptives/coherence")

    with pytest.raises(ValueError) as e:
        doc = nlp(text)  # noqa F841
    assert "Sentence vectors are not available" in str(e.value)


def test_no_warnings(nlp):
    # check that it does not raise a warning
    with warnings.catch_warnings():
        warnings.simplefilter("error")  # raise error is warning is raised
        doc = nlp("hello there!")  # noqa F841



from pathlib import Path
from typing import Iterable, List, Optional, Type, Union

import pandas as pd
import spacy
from spacy.language import Language
from spacy.tokens import Doc, Span, Token
from wasabi import msg


def get_valid_metrics() -> set:
    """Get valid metrics for extractor.

    Returns:
        set: Set of valid metrics
    """
    # extract textdescriptive components from the list of spacy Language factory
    return {
        k.split("/")[1]
        for k in Language.factories.keys()
        if k.startswith("textdescriptives")
    }


def get_doc_assigns(metric: str) -> List[str]:
    """Get doc extension attributes for a given metric.

    Args:
        metric (str): Metric to get columns for
    """
    # extract the assign names from the factory meta (this assumes that the doc._.
    # only includes elements which are also extracted as a part of the dataframe.
    if metric == "all":
        return [
            col[6:]
            for component in get_valid_metrics()
            for col in Language.get_factory_meta(
                f"textdescriptives/{component}",
            ).assigns
            if col.startswith("doc._.")
        ]
    return [
        col[6:]
        for col in Language.get_factory_meta(f"textdescriptives/{metric}").assigns
        if col.startswith("doc._.")
    ]


def get_span_assigns(metric: str) -> List[str]:
    """Get span extension attributes for a given metric.

    Args:
        metric (str): Metric to get columns for
    """
    return [
        col[7:]
        for col in Language.get_factory_meta(f"textdescriptives/{metric}").assigns
        if col.startswith("span._.")
    ]


def get_token_assigns(metric: str) -> List[str]:
    """Get token extension attributes for a given metric.

    Args:
        metric (str): Metric to get columns for
    """
    return [
        col[8:]
        for col in Language.get_factory_meta(f"textdescriptives/{metric}").assigns
        if col.startswith("token._.")
    ]


def load_sms_data():
    """Load the sms dataset
    https://archive.ics.uci.edu/ml/datasets/SMS+Spam+Collection."""
    path = Path(__file__).parent / "data" / "spam.csv"
    spam_data = pd.read_csv(path, encoding="latin-1")
    spam_data.dropna(how="any", inplace=True, axis=1)
    spam_data.columns = ["label", "message"]
    return spam_data


def _download_spacy_model(lang: str, size: str) -> str:
    """Download a spacy model for a given language and size.

    Args:
        lang (str): Language to download a model for.
        size (str): Size of the model to download. One of "sm", "md", "lg", "trf".

    Returns:
        str: Name of the downloaded model.
    """
    data_source = "news" if lang != "en" else "web"
    spacy_model = f"{lang}_core_{data_source}_{size}"
    # don't download if model already exists
    if spacy_model in spacy.cli.info()["pipelines"]:  # type: ignore
        return spacy_model
    spacy.cli.download(spacy_model)
    return spacy_model


def _remove_spacy_extension(
    spacy_language: Union[Type[Doc], Type[Span], Type[Token]],
    extension: str,
) -> None:
    """Remove spacy extension from a Language object if it exists."""
    if spacy_language.has_extension(extension):
        spacy_language.remove_extension(extension)


def _remove_textdescriptives_extensions() -> None:
    """Remove spacy extensions added by textdescriptives.

    This is necessary to avoid errors if running `extract_metrics` multiple
    times with different metrics
    """
    for metric in get_valid_metrics():
        doc_assigns = get_doc_assigns(metric)
        for assigned in doc_assigns:
            _remove_spacy_extension(spacy_language=Doc, extension=assigned)
        span_assigns = get_span_assigns(metric)
        for assigned in span_assigns:
            _remove_spacy_extension(spacy_language=Span, extension=assigned)
        token_assings = get_token_assigns(metric)
        for assigned in token_assings:
            _remove_spacy_extension(spacy_language=Token, extension=assigned)


def _create_spacy_pipeline(
    spacy_model: Optional[str],
    lang: Optional[str],
    metrics: Iterable[str],
    spacy_model_size: str,
) -> Language:
    """Load a spacy pipeline suitable for the metrics to be extracted.

    Args:
        spacy_model (Optional[str]): The spacy model to use. If not set,
            will download one based on lang.
        lang (Optional[str]): Language of the text. If lang is set and no spacy
            model is provided, will automatically download and use a spacy
            model for the language.
        metrics (Iterable[str]): Which metrics to extract.
        spacy_model_size (str): Size of the spacy model to download.

    Returns:
        Language: a spacy pipeline
    """

    metrics_requiring_spacy_model = {
        "dependency_distance",
        "pos_proportions",
        "coherence",
        "pos_proportions",
    }
    # if no spacy model is necessary for the metrics, return a blank model
    # for the language
    if not bool(metrics_requiring_spacy_model.intersection(metrics)):
        # always load spacy model if specified (if e.g. custom tokenizer)
        if spacy_model:
            return spacy.load(spacy_model)
        if lang is not None:
            return spacy.blank(lang)
        raise ValueError(
            "No spacy model needed for the metrics and no language provided. "
            + "Please provide a language to create a blank spacy model.",
        )
    # if no spacy model is provided, download one
    if spacy_model is None:
        if lang is None:
            raise ValueError(
                "No spacy model provided and no language provided. "
                + "Please provide a language to download a spacy model.",
            )
        msg.info(f"No spacy model provided. Inferring spacy model for {lang}.")
        spacy_model = _download_spacy_model(lang=lang, size=spacy_model_size)
    try:
        return spacy.load(spacy_model)
    except OSError:
        msg.info(
            f"""The specified spaCy model "{spacy_model}" was not 
            found on disk. Downloading...""",
        )
        spacy.cli.download(spacy_model)
        return spacy.load(spacy_model)



""" Calculation of statistics that require a pos-tagger in the pipeline."""

from typing import Callable, Counter, List, Union

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc, Span

from textdescriptives.components.utils import all_upos_tags


class POSProportions:
    """spaCy v.3.0 component that adds attributes for POS statistics to `Doc`
    and `Span` objects."""

    def __init__(self, nlp: Language, use_pos: bool, add_all_tags: bool):
        """Initialise components.

        Args:
            use_pos: If True, uses the simple POS tag. If False, uses the detailed
                universal POS tag.
            add_all_tags: If True, returns proportions of all possible POS tags.
                If False, only returns proportions for the POS tags present in the
                text.
        """
        self.use_pos: bool = use_pos
        self.add_all_tags: bool = add_all_tags
        self.model_tags: List[str] = (
            all_upos_tags if use_pos else nlp.meta["labels"]["tagger"]
        )

        if not Doc.has_extension("pos_proportions"):
            Doc.set_extension("pos_proportions", getter=self.pos_proportions)

        if not Span.has_extension("pos_proportions"):
            Span.set_extension("pos_proportions", getter=self.pos_proportions)

    def pos_proportions(self, text: Union[Doc, Span]) -> dict:
        """Calculates the proportion of tokens in a `Doc`|`Span` that are tagged
        with each POS tag.

        Returns:
            Dict containing {pos_prop_POSTAG: proportion of all tokens tagged with
                POSTAG.
        """
        pos_counts: Counter = Counter()
        if self.add_all_tags:
            # add all tags to the counter so they are included in the output
            pos_counts.update(self.model_tags)
            # reset all counts to 0
            pos_counts.subtract(self.model_tags)

        if self.use_pos:
            pos_counts.update([token.pos_ for token in text])
        else:
            pos_counts.update([token.tag_ for token in text])

        if self.add_all_tags:
            # filter out tags that are not in self.model_tags
            pos_counts = {  # type: ignore
                tag: count
                for tag, count in pos_counts.items()
                if tag in self.model_tags
            }

        len_text = len(text)
        return {
            f"pos_prop_{tag}": count / len(text) if len_text > 0 else np.nan
            for tag, count in pos_counts.items()
        }

    def __call__(self, doc):
        """Run the pipeline component."""
        return doc


@Language.factory(
    "textdescriptives/pos_proportions",
    assigns=["doc._.pos_proportions", "span._.pos_proportions"],
    default_config={"use_pos": True, "add_all_tags": True},
)
def create_pos_proportions_component(
    nlp: Language,
    name: str,
    use_pos: bool,
    add_all_tags: bool,
) -> Callable[[Doc], Doc]:
    """Allows PosPropotions to be added to a spaCy pipe using
    nlp.add_pipe("textdescriptives/pos_proportions")

    Adding this component to a pipeline sets the following attributes:
        - `doc._.pos_proportions`
        - `span._.pos_proportions`

    Args:
        nlp (Language): spaCy language object, does not need to be specified in the
            nlp.add_pipe call.
        name (str): name of the component. Can be optionally specified in the
            nlp.add_pipe call, using the name argument.
        use_pos: If True, uses the simple token.pos attribute. If False, uses the
            detailed token.tag attribute.

    Returns:
        Callable[[Doc], Doc]: The POSProportions component to be added to the pipe.

    Example:
        >>> import spacy
        >>> import textdescriptives as td
        >>> nlp = spacy.load("en_core_web_sm")
        >>> nlp.add_pipe("textdescriptives/pos_proportions")
        >>> # apply the component to a document
        >>> doc = nlp("This is a test sentence.")
        >>> doc._.pos_proportions
    """

    tagger = {"tagger", "attribute_ruler"}
    if not tagger.intersection(set(nlp.pipe_names)):
        raise ValueError(
            "The pipeline does not contain a component for POS tagging. Please load "
            + "a spaCy model which includes a 'tagger' or an 'attribute ruler' "
            + "component.",
        )
    return POSProportions(nlp, use_pos=use_pos, add_all_tags=add_all_tags)



""" Calculation of various readability metrics."""

from typing import Callable, Dict

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc
from wasabi import msg

from .descriptive_stats import (  # noqa
    create_descriptive_stats_component,
    language_exists_in_pyphen,
)
from .utils import filter_tokens


class Readability:
    """spaCy v.3.0 component for adding readability metrics to `Doc` objects.

    Extracts metrics and returns them as a dictionary as the ._.readability
    attribute.
    """

    def __init__(self, nlp: Language):
        """Initialise components."""
        self.can_calculate_syllables = language_exists_in_pyphen(lang=nlp.lang)

        if not Doc.has_extension("readability"):
            Doc.set_extension("readability", getter=self.readability)

    def _flesch_reading_ease(self, doc: Doc):
        """Calculate the Flesch Reading Ease score for a document. The equation
        for the Flesch Reading Ease score is:

        206.835 - (1.015 X avg sent len) - (84.6 * avg_syl_per_word)

        Higher = easier to read
        """
        if not self.can_calculate_syllables:
            return np.nan

        avg_sentence_length = doc._.sentence_length["sentence_length_mean"]
        avg_syl_per_word = doc._.syllables["syllables_per_token_mean"]
        if avg_sentence_length == 0 or avg_syl_per_word == 0:
            return np.nan
        score = 206.835 - (1.015 * avg_sentence_length) - (84.6 * avg_syl_per_word)
        return score

    def _flesch_kincaid_grade(self, doc: Doc):
        """Calculate the Flesch-Kincaid grade of the document. The equation for
        the Flesch-Kincaid grade is:

        0.39 * (avg sent len) + 11.8 * (avg_syl_per_word) - 15.59
        """
        if not self.can_calculate_syllables:
            return np.nan

        avg_sentence_length = doc._.sentence_length["sentence_length_mean"]
        avg_syl_per_word = doc._.syllables["syllables_per_token_mean"]
        if avg_sentence_length == 0 or avg_syl_per_word == 0:
            return np.nan
        score = 0.39 * avg_sentence_length + 11.8 * avg_syl_per_word - 15.59
        return score

    def _smog(self, doc: Doc, n_hard_words: int):
        """Calculate the SMOG index of the document. The equation for the SMOG index
        is:

        1.043( sqrt(30 * (hard words /n sentences)) + 3.1291

        Where hard words are words with 3 or more syllables.
        Preferably need 30+ sentences. Will not work with less than 4
        """
        if not self.can_calculate_syllables:
            return np.nan

        n_sentences = doc._._n_sentences
        if n_sentences >= 3:
            smog = (1.043 * (30 * (n_hard_words / n_sentences)) ** 0.5) + 3.1291
            return smog
        return np.nan

    def _gunning_fog(self, doc, n_hard_words: int):
        """Calculates the Gunning Fog index of the document. The equation for
        the Gunning Fog index is:

        Grade level = 0.4 * ((avg_sentence_length) + (percentage hard words))

        Where hard words are word with 3 or more syllables.
        """
        if not self.can_calculate_syllables:
            return np.nan

        n_tokens = doc._._n_tokens
        if n_tokens == 0:
            return np.nan
        avg_sent_len = doc._.sentence_length["sentence_length_mean"]
        percent_hard_words = (n_hard_words / n_tokens) * 100
        return 0.4 * (avg_sent_len + percent_hard_words)

    def _automated_readability_index(self, doc: Doc):
        """Calculates the Automated Readability Index of the document. The
        equation for the Automated Readability Index is:

        4.71 * (n_chars / n_words) + 0.5 * (n_words / n_sentences) - 21.43

        Score = grade required to read the text
        """
        if len(doc) == 0:
            return np.nan
        score = (
            4.71 * doc._.token_length["token_length_mean"]
            + 0.5 * doc._.sentence_length["sentence_length_mean"]
            - 21.43
        )
        return score

    def _coleman_liau_index(self, doc: Doc):
        """Calculates the Colmean-Liau index of the document. The equation for
        the Coleman-Liau index is:

        score = 0.0588 * avg number of chars pr 100 words - 0.296 * avg num of sents pr
        100 words -15.8

        Score = grade required to read the text
        """
        n_tokens = doc._._n_tokens
        if n_tokens == 0:
            return np.nan
        lengths = doc._.token_length["token_length_mean"] * 100
        s = (doc._._n_sentences / n_tokens) * 100
        return 0.0588 * lengths - 0.296 * s - 15.8

    def _lix(self, doc: Doc, long_words: int):
        """Calculates the LIX index of the document. The equation for the LIX index
        is:

        (n_words / n_sentences) + (n_words longer than 6 characters * 100) / n_words
        """
        n_tokens = doc._._n_tokens
        if n_tokens == 0:
            return np.nan
        percent_long_words = long_words / n_tokens * 100
        return doc._.sentence_length["sentence_length_mean"] + percent_long_words

    def _rix(self, doc: Doc, long_words: int):
        """Calculates the RIX index of the document. The equation for the RIX index
        is:

        (n_words longer than 6 characters / n_sentences)
        """
        n_sentences = doc._._n_sentences
        if doc._._n_tokens == 0:
            return np.nan
        return long_words / n_sentences

    def readability(self, doc: Doc) -> Dict[str, float]:
        """Apply readability functions and return a dict of the results."""
        hard_words = (
            len([syllable for syllable in doc._._n_syllables if syllable >= 3])
            if self.can_calculate_syllables
            else 0
        )
        long_words = len([t for t in filter_tokens(doc) if len(t) > 6])

        return {
            "flesch_reading_ease": self._flesch_reading_ease(doc),
            "flesch_kincaid_grade": self._flesch_kincaid_grade(doc),
            "smog": self._smog(doc, hard_words),
            "gunning_fog": self._gunning_fog(doc, hard_words),
            "automated_readability_index": self._automated_readability_index(doc),
            "coleman_liau_index": self._coleman_liau_index(doc),
            "lix": self._lix(doc, long_words),
            "rix": self._rix(doc, long_words),
        }

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        return doc


@Language.factory(
    "textdescriptives/readability",
    assigns=["doc._.readability"],
    default_config={"verbose": False},
)
def create_readability_component(
    nlp: Language,
    name: str,
    verbose: bool,
) -> Callable[[Doc], Doc]:
    """Allows Readability to be added to a spaCy pipe using
    nlp.add_pipe("textdescriptives/readability").

    Readability requires attributes from DescriptiveStatistics and adds
    it to the pipe if it not already loaded.

    Adding this component to a pipeline sets the following attributes:
        - doc._.readability

    Args:
        nlp (Language): spaCy language object, does not need to be specified in the
            nlp.add_pipe call.
        name (str): name of the component. Can be optionally specified in the
            nlp.add_pipe call, using the name argument.
        verbose (bool): Toggle to show a message if the
            "textdescriptives/descriptive_stats" component is added to the pipeline.
            Defaults to True.

    Returns:
        Callable[[Doc], Doc]: The Readability component

    Example:
        >>> import spacy
        >>> import textdescriptives as td
        >>> nlp = spacy.blank("en")
        >>> nlp.add_pipe("textdescriptives/readability")
        >>> # apply the pipeline to a document
        >>> doc = nlp("This is a test document.")
        >>> doc._.readability
    """
    if "textdescriptives/descriptive_stats" not in nlp.pipe_names:
        if verbose:
            msg.info(  # pylint: disable=logging-not-lazy
                "'textdescriptives/descriptive_stats' component is required for"
                + " 'textdescriptives.readability'. Adding to pipe.",
            )
        nlp.add_pipe("textdescriptives/descriptive_stats")
    return Readability(nlp)



""" Calculation of descriptive statistics."""

from typing import Callable, Dict, Union

import numpy as np
from pyphen import Pyphen
from spacy.language import Language
from spacy.tokens import Doc, Span
from wasabi import msg

from .utils import filter_tokens, n_sentences, n_syllables, n_tokens


def language_exists_in_pyphen(lang: str) -> bool:
    try:
        _ = Pyphen(lang=lang)
        return True
    except KeyError:
        return False


class DescriptiveStatistics:
    """spaCy v.3.0 component that adds attributes with descriptive statistics to
    `Doc` and `Span` objects.

    The attributes relate to token and sentence length, number of syllables, and
    counts of tokens and sentences.
    """

    def __init__(self, nlp: Language, verbose: bool):
        """Initialise components."""
        self.can_calculate_syllables = language_exists_in_pyphen(lang=nlp.lang)
        if not self.can_calculate_syllables and verbose:
            msg.warn(
                f"Could not load syllable counter for language {nlp.lang}. "
                + "The following extensions will be set to np.nan: "
                + "syllables, flesch_reading_ease, flesch_kincaid_grade, "
                + "smog, gunning_fog.",
            )

        extensions: Dict[str, Callable] = {
            "_n_sentences": n_sentences,
            "_n_tokens": n_tokens,
            "_n_syllables": n_syllables,
            "token_length": self.token_length,
            "sentence_length": self.sentence_length,
            "syllables": self.syllables,
            "counts": self.counts,
            "descriptive_stats": self.descriptive_stats,
        }

        for extension_name, getter_fun in extensions.items():
            if extension_name not in [
                "_n_sentences",
                "sentence_length",
                "syllables",
            ] and not Span.has_extension(extension_name):
                Span.set_extension(extension_name, getter=getter_fun)
            if not Doc.has_extension(extension_name):
                Doc.set_extension(extension_name, getter=getter_fun)

    def token_length(self, doc: Union[Doc, Span]) -> dict:
        """Calculate mean, median and std of token length for a `Doc` or `Span`.

        Returns:
            dict: token_length_mean, token_length_median, token_length_std
        """
        token_lengths = [len(token) for token in filter_tokens(doc)]
        if not token_lengths:
            return {
                "token_length_mean": np.nan,
                "token_length_median": np.nan,
                "token_length_std": np.nan,
            }
        return {
            "token_length_mean": np.mean(token_lengths),
            "token_length_median": np.median(token_lengths),
            "token_length_std": np.std(token_lengths),
        }

    def sentence_length(self, doc: Doc) -> dict:
        """Calculate mean, median and std of sentence length for a `Doc`.

        Returns:
            dict: sentence_length_mean, sentence_length_median, sentence_length_std
        """
        # get length of filtered tokens per sentence
        tokenized_sentences = [
            [
                token.text
                for token in sent
                if not token.is_punct and "'" not in token.text
            ]
            for sent in doc.sents
        ]
        len_sentences = [len(sentence) for sentence in tokenized_sentences]
        if not len_sentences:
            return {
                "sentence_length_mean": np.nan,
                "sentence_length_median": np.nan,
                "sentence_length_std": np.nan,
            }
        return {
            "sentence_length_mean": np.mean(len_sentences),
            "sentence_length_median": np.median(len_sentences),
            "sentence_length_std": np.std(len_sentences),
        }

    def syllables(self, doc: Doc) -> dict:
        """Calculate mean, median and std of syllables per token for a `Doc`.
        Uses `Pyphen` for hyphenation.

        Returns:
            dict: syllables_per_token_mean, syllables_per_token_median,
                syllables_per_token_std
        """
        nan_output = {
            "syllables_per_token_mean": np.nan,
            "syllables_per_token_median": np.nan,
            "syllables_per_token_std": np.nan,
        }

        if not self.can_calculate_syllables:
            return nan_output
        if n_syllables := doc._._n_syllables:
            return {
                "syllables_per_token_mean": np.mean(n_syllables),
                "syllables_per_token_median": np.median(n_syllables),
                "syllables_per_token_std": np.std(n_syllables),
            }
        else:
            return nan_output

    def counts(self, doc: Union[Doc, Span], ignore_whitespace: bool = True) -> dict:
        """Calculate counts of tokens, unique tokens, and characters for a `Doc`
        or `Span`. Adds number of sentences for `Doc` objects.

        Args:
            ignore_whitespace: if True, whitespace is not counted as a character when
                counting number of characters.
        Return:
            dict: n_tokens, n_unique_tokens, proportion_unique_tokens, n_characters,
                (n_sentences)
        """
        n_tokens = doc._._n_tokens
        n_types = len({tok.lower_ for tok in filter_tokens(doc)})
        if ignore_whitespace:
            n_chars = len(doc.text.replace(" ", ""))
        else:
            n_chars = len(doc.text)

        prop_unique_tokens = np.nan if n_tokens == 0 else n_types / n_tokens
        out = {
            "n_tokens": n_tokens,
            "n_unique_tokens": n_types,
            "proportion_unique_tokens": prop_unique_tokens,
            "n_characters": n_chars,
        }
        if isinstance(doc, Doc):
            out["n_sentences"] = doc._._n_sentences
        return out

    def descriptive_stats(self, doc: Union[Doc, Span]) -> dict:
        """Get all descriptive statistics in a single dict."""
        out = {**doc._.counts, **doc._.token_length}
        if isinstance(doc, Span):
            return out
        return {**out, **doc._.sentence_length, **doc._.syllables}

    def __call__(self, doc):
        """Run the pipeline component."""
        return doc


@Language.factory(
    "textdescriptives/descriptive_stats",
    assigns=[
        "doc._._n_sentences",
        "doc._._n_tokens",
        "doc._._n_syllables",
        "doc._.token_length",
        "doc._.sentence_length",
        "doc._.syllables",
        "doc._.counts",
        "doc._.descriptive_stats",
        "span._._n_tokens",
        "span._._n_syllables",
        "span._.token_length",
        "span._.counts",
        "span._.descriptive_stats",
    ],
    default_config={"verbose": True},
)
def create_descriptive_stats_component(
    nlp: Language,
    name: str,
    verbose: bool,
) -> Callable[[Doc], Doc]:
    """Allows DescriptiveStatistics to be added to a spaCy pipe using
    nlp.add_pipe("textdescriptives/descriptive_stats").

    Adding the component to the pipe will add the following attributes to
    `Doc` and `Span` objects:

    - `doc._.n_sentences`
    - `doc._.n_tokens`
    - `doc._.token_length`
    - `doc._.sentence_length`
    - `doc._.syllables`
    - `doc._.counts`
    - `doc._.descriptive_stats`
    - `span._.token_length`
    - `span._.counts`
    - `span._.descriptive_stats`

    Args:
        nlp (Language): spaCy language object, does not need to be specified in the
            nlp.add_pipe call.
        name (str): name of the component. Can be optionally specified in the
            nlp.add_pipe call, using the name argument.

    Returns:
        Callable[[Doc], Doc]: DescriptiveStatistics component

    Example:
        >>> import spacy
        >>> import textdescriptives as td
        >>> nlp = spacy.blank("en")
        >>> # add sentencizer
        >>> nlp.add_pipe("sentencizer")
        >>> # add descriptive stats
        >>> nlp.add_pipe("textdescriptives/descriptive_stats")
        >>> # apply to a document
        >>> doc = nlp("This is a sentence. This is another sentence.")
        >>> doc._.descriptive_stats
    """
    sentencizers = {"sentencizer", "parser"}
    if not sentencizers.intersection(set(nlp.pipe_names)):
        nlp.add_pipe("sentencizer")  # add a sentencizer if not one in pipe
    return DescriptiveStatistics(nlp, verbose=verbose)



from .about import __title__, __version__  # noqa: F401
from .components.quality import QualityThresholds  # noqa: F401
from .extractors import extract_df, extract_dict, extract_metrics  # noqa: F401
from .load_components import TextDescriptives  # noqa: F401
from .utils import get_doc_assigns, get_valid_metrics  # noqa: F401



import pytest
import spacy
import textdescriptives as td  # noqa: F401


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives/all")
    return nlp


def test_integration(nlp):
    assert nlp.pipe_names[-1] == "textdescriptives/all"
    for component in [
        "textdescriptives/descriptive_stats",
        "textdescriptives/readability",
        "textdescriptives/dependency_distance",
        "textdescriptives/all",
        "textdescriptives/quality",
        "textdescriptives/pos_proportions",
    ]:
        assert component in nlp.pipe_names


def test_simple(nlp):
    doc = nlp("This is a short and simple text")
    assert doc



""" Data classes used for the quality component."""

from typing import Any, Dict, Optional, Tuple, Union

from pydantic import ConfigDict, BaseModel, Field

Interval = Tuple[Optional[float], Optional[float]]


class ThresholdsOutput(BaseModel):
    """An output which contains an three items. 1) a thresholds which is either
    an interval or a accepted boolean value. 2) a value which is the value of
    the metric. 3) a boolean which is True if the value is within the
    thresholds.

    Example:
        >>> t_out = ThresholdsOutput(threshold=(0, 2), value=2)
        >>> t_out
        ThresholdsOutput(value=2.0, passed=True, threshold=(0.0, 2.0))
        >>> t_out.passed
        True
    """

    model_config = ConfigDict(extra="forbid")

    threshold: Union[Interval, bool, None]
    value: Union[float, None]

    @property
    def passed(self) -> Optional[bool]:
        """Return True if the value is within the thresholds."""
        if self.value is None:
            return None
        if self.threshold is None:
            return True
        if isinstance(self.threshold, bool):
            return self.threshold == self.value
        lower, upper = self.threshold
        return (lower is None or lower <= self.value) and (
            upper is None or self.value <= upper
        )

    def __repr_str__(self, join_str: str) -> str:
        value = round(self.value, 2) if isinstance(self.value, float) else self.value
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}"
            for a, v in [
                ("value", value),
                ("passed", self.passed),
                ("threshold", self.threshold),
            ]
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ThresholdsOutput):
            return self.value == other.value and self.threshold == other.threshold
        return self.value == other


class QualityThresholds(BaseModel):
    """Thresholds for quality metrics."""

    model_config = ConfigDict(extra="forbid")

    n_stop_words: Interval = Field(
        (2, None),
        description="A Range for the number of stop words. Default: (2, None), i.e. "
        + "at least 2 stop words, but no upper limit.",
    )
    alpha_ratio: Interval = Field(
        (0.7, None),
        description="A Range for the alpha ratio. Default: (0.7, None), i.e. at "
        + r"least 70% of tokens contain at least one alphabetic character, but no "
        + "upper limit. Note this is lowered from the original 0.8 to account for a"
        + "different definition of word boundaries. E.g. in spaCy a punctuation is"
        + "not a part of a word.",
    )
    mean_word_length: Interval = Field(
        (3, 10),
        description="A Range for the mean word length. Default: (3, 10), i.e. between"
        + " 3 and 10 characters.",
    )
    doc_length: Interval = Field(
        (10, 100_000),
        description="A Range for the document length. Default: (10, 100_000), i.e."
        + " between 10 and 100_000 characters.",
    )
    symbol_to_word_ratio: Dict[str, Interval] = Field(
        {"#": (None, 0.1)},
        description="A dict of symbols and the allowed range for the "
        + r"symbol-to-word-ratio. The symbol-to-word-ratio is the ratio between symbol"
        + "occurrence and word occurrence. Defaults to {'#': (None, 0.1)} i.e. no lower"
        + r" limit, but there must at most be a ratio of 0.1 between the number of of "
        + "words and hashtags. i.e. if we have 100 words the symbol should appear no "
        + "more than 10 times. Values not in the dict are not checked.",
    )
    proportion_ellipsis: Interval = Field(
        (None, 0.3),
        description="A Range for the proportion of lines which end with ellipsis. "
        + "Default: (None, 0.3), "
        + r"i.e. no lower limit, but at most 30% of lines end with an ellipsis.",
    )
    proportion_bullet_points: Interval = Field(
        (None, 0.8),
        description="A Range for the proportion lines which start with a bullet "
        + r"points. Default: (None, 0.8), i.e. no lower limit, but at most 80% of lines"
        + " start with a bullet point.",
    )
    contains: Dict[str, bool] = Field(
        {"lorem ipsum": False},
        description="A dictionary of strings and whether they should be contained in "
        + "the document. Default: {'lorem ipsum': False}, i.e. the document should not"
        + " contain the string 'lorem ipsum'.",
    )
    duplicate_line_chr_fraction: Interval = Field(
        (None, 0.2),
        description="A Range for the duplicate line character fraction. Default: "
        + r"(None, 0.2), i.e. no lower limit, but at most 20% of characters are"
        + " duplicates.",
    )
    duplicate_paragraph_chr_fraction: Interval = Field(
        (None, 0.2),
        description="A Range for the duplicate paragraph character fraction. Default:"
        + r" (None, 0.2), i.e. no lower limit, but at most 20% of characters are "
        + "duplicates.",
    )
    duplicate_ngram_chr_fraction: Dict[str, Interval] = Field(
        {
            "5": (None, 0.15),
            "6": (None, 0.14),
            "7": (None, 0.13),
            "8": (None, 0.12),
            "9": (None, 0.11),
            "10": (None, 0.1),
        },
        description="A dictionary of n-gram lengths and the allowed range for the "
        + "duplicate n-gram character fraction. Default: {5: (None, 0.15), 6: (None, "
        + "0.14), 7: (None, 0.13), 8: (None, 0.12), 9: (None, 0.11), 10: (None, 0.1)}, "
        + r"i.e. no lower limit, but at most 15% of characters are duplicates for "
        + r"5-grams, 14% for 6-grams, 13% for 7-grams, 12% for 8-grams, 11% for 9-grams"
        + r" and 10% for 10-grams.",
    )
    top_ngram_chr_fraction: Dict[str, Interval] = Field(
        {
            "2": (None, 0.2),
            "3": (None, 0.18),
            "4": (None, 0.16),
        },
        description="A dictionary of n-gram lengths and the allowed range for the "
        + "top n-gram character fraction. Default: {2: (None, 0.2), 3: (None, 0.18)"
        + r", 4: (None, 0.16)}, i.e. no lower limit, but at most 20% of characters "
        + r"are contained within a duplicate for 2-grams, 18% for 3-grams and 16% "
        + "for 4-grams.",
    )
    oov_ratio: Interval = Field(
        (None, 0.2),
        description="A range for the out-of-vocabulary ratio. Default: (None, 0.2)"
        + r" i.e. no lower limit, but at most 20% of words are out-of-vocabulary.",
    )


class QualityOutput(BaseModel):
    """The output of the quality function."""

    model_config = ConfigDict(extra="forbid")

    n_stop_words: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the number of stop words.",
    )
    alpha_ratio: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the alpha ratio.",
    )
    mean_word_length: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the mean word length.",
    )
    doc_length: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the document length.",
    )
    symbol_to_word_ratio: Dict[str, ThresholdsOutput] = Field(
        ...,
        description="The thresholds output for the symbol-to-word-ratio.",
    )
    proportion_ellipsis: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the proportion of lines ending with "
        + "ellipsis.",
    )
    proportion_bullet_points: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the proportion of lines starting with "
        + "bullet points.",
    )
    contains: Dict[str, ThresholdsOutput] = Field(
        ...,
        description="The thresholds output for the presence of strings.",
    )
    duplicate_line_chr_fraction: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the duplicate line character fraction.",
    )
    duplicate_paragraph_chr_fraction: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the duplicate paragraph character "
        + "fraction.",
    )
    duplicate_ngram_chr_fraction: Dict[str, ThresholdsOutput] = Field(
        ...,
        description="The thresholds output for the duplicate n-gram character "
        + "fraction.",
    )
    top_ngram_chr_fraction: Dict[str, ThresholdsOutput] = Field(
        ...,
        description="The thresholds output for the top n-gram character fraction.",
    )
    oov_ratio: ThresholdsOutput = Field(
        ...,
        description="The thresholds output for the out-of-vocabulary ratio.",
    )

    @property
    def passed(self) -> bool:
        """
        Returns:
            bool: Whether all thresholds have been passed.
        """
        passed_or_none = [
            self.n_stop_words.passed,
            self.alpha_ratio.passed,
            self.mean_word_length.passed,
            self.doc_length.passed,
            all(v.passed for v in self.symbol_to_word_ratio.values()),
            self.proportion_ellipsis.passed,
            self.proportion_bullet_points.passed,
            all(v.passed for v in self.contains.values()),
            self.duplicate_line_chr_fraction.passed,
            self.duplicate_paragraph_chr_fraction.passed,
            all(v.passed for v in self.duplicate_ngram_chr_fraction.values()),
            all(v.passed for v in self.top_ngram_chr_fraction.values()),
            self.oov_ratio.passed,
        ]

        return all(i is None or i for i in passed_or_none)

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"\n\t{a}={v!r}"
            for a, v in [
                ("passed", self.passed),
            ]
            + list(self.__repr_args__())
        )

    def to_flat_value_dict(self) -> Dict[str, Any]:
        """Creates a flat dictionary representation of the object to allow for
        easy easy conversion to a pandas DataFrame."""
        flat_dict = {"passed_quality_check": self.passed}

        for k, v in self.__dict__.items():
            if isinstance(v, dict):
                for k2, v2 in v.items():
                    flat_dict[f"{k}_{k2}"] = v2.value
            else:
                flat_dict[k] = v.value

        return flat_dict



from typing import Callable, List

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc


def n_order_coherence(doc: Doc, order: int) -> List[float]:
    """Calculate coherence for a `Doc` for a given order.

    Args:
        doc: A `Doc` object.
        order: The order of coherence to calculate. For example, order=1 will
            calculate the semantic similarity between consecutive sentences. And
            order=2 will calculate the semantic similarity between sentences that
            are two sentences apart.

    Returns:
        A list of floats representing the semantic similarity between sentences
    """

    if not doc.has_annotation("SENT_START"):
        raise ValueError(
            "A sentence boundary detector has not been run on this Doc, which is "
            + "required to calculate coherence. Have you added a model with a "
            + "sentencizer and word vectors to the pipeline?",
        )
    sents = list(doc.sents)
    if len(sents) < order + 1:
        return [np.nan]

    if doc.vector.size == 0:
        raise ValueError(
            "Sentence vectors are not available. Thus it is not possible to "
            + "calculate the coherence between sentences. Please add a component "
            + "that includes word vectors or sentence embeddings."
            + "See https://spacy.io/usage/vectors-similarity for more details.",
        )

    similarities: List[float] = []
    for i, sent in enumerate(sents):
        if i == len(sents) - order:
            break
        similarities.append(sent.similarity(sents[i + order]))
    return similarities


class Coherence:
    """Spacy v.3.0 component that adds attributes with coherence to `Doc` and
    `Span` objects."""

    def __init__(self, nlp: Language):
        """Initialise component."""
        extensions = [
            "first_order_coherence_values",
            "second_order_coherence_values",
            "coherence",
        ]
        for extension in extensions:
            if not Doc.has_extension(extension):
                Doc.set_extension(extension, default=None)

    @staticmethod
    def _first_order_coherence(doc: Doc) -> List[float]:
        """Calculate first order coherence for a `Doc`, i.e. the semantic
        similarity between consecutive sentences."""
        return n_order_coherence(doc=doc, order=1)

    @staticmethod
    def _second_order_coherence(doc: Doc) -> List[float]:
        """Calculate second order coherence for a `Doc`, i.e. the semantic
        similarity between sentences that are two sentences apart."""
        return n_order_coherence(doc, order=2)

    def coherence(self, doc: Doc) -> None:
        """Calculate mean semantic coherence for a `Doc` and set the coherence
        attribute.

        Coherence is calculated by taking the mean of the similarity between
        sentence embeddings. See the documentation for more details.
        """
        first_order_coherence = self._first_order_coherence(doc)
        second_order_coherence = self._second_order_coherence(doc)

        # get mean of coherence values
        if len(first_order_coherence) < 2:
            first_order_coherence_mean = first_order_coherence[0]
        else:
            first_order_coherence_mean = np.nanmean(first_order_coherence)
        if len(second_order_coherence) < 2:
            second_order_coherence_mean = second_order_coherence[0]
        else:
            second_order_coherence_mean = np.nanmean(second_order_coherence)

        # set attributes
        setattr(doc._, "first_order_coherence_values", first_order_coherence)
        setattr(doc._, "second_order_coherence_values", second_order_coherence)
        setattr(
            doc._,
            "coherence",
            {
                "first_order_coherence": first_order_coherence_mean,
                "second_order_coherence": second_order_coherence_mean,
            },
        )

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        self.coherence(doc)
        return doc


@Language.factory(
    "textdescriptives/coherence",
    assigns=[
        "doc._.first_order_coherence_values",
        "doc._.second_order_coherence_values",
        "doc._.coherence",
    ],
)
def create_coherence_component(nlp: Language, name: str) -> Callable[[Doc], Doc]:
    """Allows Coherence to be added to a spaCy pipe using
    nlp.add_pipe("textdescriptives/coherence").

    Adding this component to a pipeline sets the following attributes:
        - doc._.first_order_coherence_values
        - doc._.second_order_coherence_values
        - doc._.coherence

    Args:
        nlp (Language): spaCy language object, does not need to be specified in the
            nlp.add_pipe call.
        name (str): name of the component. Can be optionally specified in the
            nlp.add_pipe call, using the name argument.

    Returns:
        Callable[[Doc], Doc]: The Coherence component to be added to the pipe.

    Examples:
        >>> import spacy
        >>> import textdescriptives as td
        >>> nlp = spacy.load("en_core_web_md")
        >>> nlp.add_pipe("textdescriptives/coherence")
        >>> # apply the pipeline to a text
        >>> doc = nlp("This is a sentence. This is another sentence.")
        >>> # get coherence values
        >>> doc._.coherence
    """
    return Coherence(nlp)



from .coherence import Coherence  # noqa: F401
from .dependency_distance import DependencyDistance  # noqa: F401
from .descriptive_stats import DescriptiveStatistics  # noqa: F401
from .pos_proportions import POSProportions  # noqa: F401
from .quality import Quality  # noqa: F401
from .readability import Readability  # noqa: F401
from .information_theory import InformationTheory  # noqa: F401



""" About textdescriptives, version number is specified in the setup.cfg
file."""

# if python >= 3.8, use importlib.metadata otherwise use pkg_resources
try:
    from importlib.metadata import version

    __version__ = version("textdescriptives")
except ImportError:
    from pkg_resources import get_distribution  # type: ignore

    __version__ = get_distribution("textdescriptives").version

__title__ = "textdescriptives"
__download_url__ = "https://github.com/HLasse/textdescriptives"



""" Utility functions for calculating various text descriptives."""

from typing import Union

from pyphen import Pyphen
from spacy.tokens import Doc, Span, Token


def filter_tokens(doc: Union[Doc, Span]):
    """Return words in document or span.

    Filters punctuation and words that start with an apostrophe (contractions)
    """
    filtered_tokens = [
        word for word in doc if not word.is_punct and "'" not in word.text
    ]
    return filtered_tokens


def n_sentences(doc: Doc):
    """Return number of sentences in the document."""
    return len(list(doc.sents))


def n_tokens(doc: Union[Doc, Span]):
    """Return number of words in the document."""
    return len(filter_tokens(doc))


def n_syllables(doc: Doc):
    """Return number of syllables per token."""
    dic = Pyphen(lang=doc.lang_)

    def count_syl(token: Token):
        word_hyphenated = dic.inserted(token.lower_)
        return max(1, word_hyphenated.count("-") + 1)

    return [count_syl(token) for token in filter_tokens(doc)]


all_upos_tags = [
    "ADJ",
    "ADP",
    "ADV",
    "AUX",
    "CCONJ",
    "DET",
    "INTJ",
    "NOUN",
    "NUM",
    "PART",
    "PRON",
    "PROPN",
    "PUNCT",
    "SCONJ",
    "SYM",
    "VERB",
    "X",
]



import numpy as np
import pytest
import spacy

import textdescriptives as td
from textdescriptives.components.information_theory import (
    entropy_getter,
    per_word_perplexity_getter,
    perplexity_getter,
    set_lexeme_prob_table,
)
from textdescriptives.utils import _remove_textdescriptives_extensions


@pytest.fixture
def nlp():
    nlp = spacy.blank("en")
    return nlp


def test_unigram_information_metrics(nlp):  # noqa F811
    set_lexeme_prob_table(nlp.vocab, verbose=True)
    doc1 = nlp("This is a very likely sentence.")
    doc2 = nlp("This is a unlikely sentence sadsatrsxss.")

    assert doc1[0].prob != -20, "lexeme_prob table not loaded."
    assert perplexity_getter(doc1) > perplexity_getter(doc2)
    assert per_word_perplexity_getter(doc1) > per_word_perplexity_getter(doc2)
    assert entropy_getter(doc1) > entropy_getter(doc2)


def test_extract_df(nlp):  # noqa F811
    _remove_textdescriptives_extensions()
    nlp.add_pipe("textdescriptives/information_theory")
    doc = nlp("This is a very likely sentence.")

    df = td.extract_df(doc)
    for col in ["perplexity", "per_word_perplexity", "entropy"]:
        assert col in df.columns


def test_language_no_lexeme_prob_table():
    nlp = spacy.blank("hr")
    nlp.add_pipe("textdescriptives/information_theory")
    doc = nlp("Ez egy magyar mondat.")

    assert np.isnan(doc._.perplexity)



import pytest
import spacy

import textdescriptives as td
from textdescriptives.utils import _create_spacy_pipeline, _download_spacy_model

# pylint: disable=missing-function-docstring


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives/all")
    return nlp


def test_extract_df_single_doc(nlp):
    doc = nlp("This is just a cute little text. Actually, it's two sentences.")
    td.extract_df(doc)
    for metric in [
        "descriptive_stats",
        "readability",
        "dependency_distance",
        "quality",
    ]:
        td.extract_df(doc, metrics=metric)


def test_extract_df_pipe(nlp):
    text = [
        "I wonder how well the function works on multiple documents",
        "Very exciting to see, don't you think?",
    ]
    docs = nlp.pipe(text)
    df = td.extract_df(docs)
    assert "lix" in df.columns
    assert "dependency_distance_mean" in df.columns
    assert "n_stop_words" in df.columns
    assert "pos_prop_VERB" in df.columns
    assert "n_tokens" in df.columns
    assert "first_order_coherence" in df.columns
    assert len(df) == 2


def test_extract_dict_single_doc(nlp):
    doc = nlp("This is just a cute little text. Actually, it's two sentences.")
    td.extract_dict(doc)
    for metric in ["descriptive_stats", "readability", "dependency_distance"]:
        dict_list = td.extract_dict(doc, metrics=metric)
        assert isinstance(dict_list, list)
        assert len(dict_list) == 1

        metrics_dict = dict_list[0]
        assert isinstance(metrics_dict, dict)
        assert metrics_dict.pop("text") == doc.text


def test_extract_df_only_pos():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives/pos_proportions")

    doc = nlp("This is just a cute little text. Actually, it's two sentences.")
    td.extract_df(doc, metrics="pos_proportions")


@pytest.mark.parametrize("lang", ["en", "da"])
def test_download_spacy_model(lang):
    _download_spacy_model(lang=lang, size="sm")


def test_load_spacy_model():
    nlp = _create_spacy_pipeline(
        spacy_model=None,
        lang="en",
        metrics=["descriptive_stats", "readability", "coherence"],
        spacy_model_size="sm",
    )
    assert "tok2vec" in nlp.pipe_names


def test_load_spacy_model_blank():
    nlp = _create_spacy_pipeline(
        spacy_model=None,
        lang="en",
        metrics=["descriptive_stats"],
        spacy_model_size="sm",
    )
    assert "tok2vec" not in nlp.pipe_names


def test_extract_single_doc():
    df = td.extract_metrics(
        "This is just a cute little text. Actually, it's two sentences.",
        spacy_model="en_core_web_sm",
        metrics="readability",
    )
    assert "lix" in df.columns


def test_extract_with_lang():
    df = td.extract_metrics(
        "This is just a cute little text. Actually, it's two sentences.",
        metrics="dependency_distance",
        lang="en",
        spacy_model_size="sm",
    )
    assert "dependency_distance_mean" in df.columns


@pytest.mark.parametrize(
    "text",
    [
        "This is just a cute little text. Actually, it's two sentences. "
        + "No, it's three",
        [
            "This is just a cute little text. Actually, it's two sentences. "
            + "No, it's three.",
            "Two documents in this bad boy. Let's see how it works.",
        ],
    ],
)
def test_extract_similar_extract_df(text):
    df = td.extract_metrics(
        text,
        spacy_model="en_core_web_sm",
        metrics="coherence",
    )

    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives/coherence")
    if isinstance(text, str):
        text = [text]
    docs = nlp.pipe(text)
    df2 = td.extract_df(docs)
    assert df.equals(df2)


def test_extract_df_then_extract_metric():
    text = [
        "This is just a cute little text. Actually, it's two sentences. "
        + "No, it's three.",
        "Two documents in this bad boy. Let's see how it works.",
    ]
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives/coherence")
    docs = nlp.pipe(text)
    td.extract_df(docs)

    td.extract_metrics(
        text,
        spacy_model="en_core_web_sm",
        metrics="quality",
    )


def test_extract_model_not_needed():
    df = td.extract_metrics(
        "This is just a cute little text. Actually, it's two sentences.",
        metrics="descriptive_stats",
        lang="en",
    )
    assert "n_tokens" in df.columns


def test_extract_metrics_twice():
    text = "Just a small test"
    td.extract_metrics(
        text,
        metrics="descriptive_stats",
        lang="en",
    )
    td.extract_metrics(
        text,
        metrics="coherence",
        lang="en",
    )


def test_extract_metrics_multiple_metrics():
    df = td.extract_metrics(
        "This is just a cute little text. Actually, it's two sentences.",
        metrics=["readability", "dependency_distance"],
        spacy_model="en_core_web_sm",
    )
    assert "lix" in df.columns
    assert "dependency_distance_mean" in df.columns


@pytest.mark.parametrize(
    "text",
    ["just a little test", ""],
)
def test_extract_metrics_all_metrics(text: str):
    df = td.extract_metrics(text=text, spacy_model="en_core_web_sm", metrics=None)
    assert "n_tokens" in df.columns



from typing import Iterable, List, Optional

import pandas as pd

try:
    from sklearn.base import BaseEstimator, TransformerMixin
except ImportError as e:
    raise ImportError(
        "Failed to import sklearn. If you want to use the sklearn integration, "
        + "please install it with `pip install scikit-learn` or install "
        + "textdescriptives with the [sklearn] extra: "
        + "pip install textdescriptives[sklearn].",
    ) from e
from wasabi import msg

from textdescriptives import extract_metrics


def get_feature_names_from_metrics_and_model(
    lang: Optional[str],
    metrics: Optional[Iterable[str]],
    spacy_model: Optional[str],
    spacy_model_size: str,
) -> List[str]:
    """Get the names of the extracted features from the specified metrics
    and model. Does this by extracting the metrics from an empty dummy text."""
    df = extract_metrics(
        text="",
        lang=lang,
        metrics=metrics,
        spacy_model=spacy_model,
        spacy_model_size=spacy_model_size,
    )
    return list(df.drop("text", axis=1).columns)


class TextDescriptivesFeaturizer(TransformerMixin, BaseEstimator):
    """Wrapper for extracting text metrics using textdescriptives and
    using it in a sklearn pipeline."""

    def __init__(
        self,
        lang: Optional[str] = None,
        metrics: Optional[Iterable[str]] = None,
        spacy_model: Optional[str] = None,
        spacy_model_size: str = "lg",
    ):
        """Initialise the transformer with arguments to
        textdescriptives.extract_metrics.

        Args:
            text (Union[str, List[str]]): A text or a list of texts.
            lang (str, optional): Language of the text. If lang is set and no spacy
                model is provided, will automatically download and use a spacy
                model for the language. Defaults to None.
            metrics (List[str]): Which metrics to extract.
                One or more of ["descriptive_stats", "readability",
                "dependency_distance", "pos_proportions", "coherence", "quality",
                "information_theory"]. If None, will extract all metrics from
                textdescriptives. Defaults to None.
            spacy_model (str, optional): The spacy model to use. If not set,
                will download one based on lang. Defaults to None.
            spacy_model_size (str, optional): Size of the spacy model to download.
        """
        self.lang = lang
        if isinstance(metrics, str):
            metrics = [metrics]
        self.metrics = metrics
        self.spacy_model = spacy_model
        self.spacy_model_size = spacy_model_size

        if spacy_model is None and lang is None:
            raise ValueError("Either a spacy model or a language must be provided.")
        if spacy_model is not None and lang is not None:
            msg.info(
                "Both a spacy model and a language were provided. "
                + "Will use the spacy model and ignore language.",
            )
        self.feature_names = get_feature_names_from_metrics_and_model(
            lang=self.lang,
            metrics=self.metrics,
            spacy_model=self.spacy_model,
            spacy_model_size=self.spacy_model_size,
        )

    def fit(self, X, y=None):
        """Fit the transformer to the data. This is not needed for this
        transformer, but is required for sklearn compatibility."""
        return self

    def transform(self, X) -> pd.DataFrame:
        """Transform the data using textdescriptives.

        Args:
            X: Iterable of strings.

        Returns:
            Numpy array of shape (n_samples, n_features).
        """
        metrics = extract_metrics(
            X,
            lang=self.lang,
            metrics=self.metrics,
            spacy_model=self.spacy_model,
            spacy_model_size=self.spacy_model_size,
        )
        return metrics.drop("text", axis=1)

    def get_feature_names(self) -> List[str]:
        """Get the names of the extracted features."""
        return self.feature_names

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """Get the names of the extracted features. input_features is only
        present for API compatibility with sklearn."""
        return self.feature_names
        return self.feature_names



""" Component for calculating quality metrics."""

from collections import Counter, defaultdict
from typing import Callable, Dict, List, Mapping, Optional, Tuple, Union

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc, Span

from .quality_data_classes import QualityOutput, QualityThresholds, ThresholdsOutput


def n_stop_words(span: Union[Doc, Span]) -> int:
    """Count the number of stop words in a document.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object

    Returns:
        int: number of stop words
    """
    return sum(t.is_stop for t in span)


def mean_word_length(span: Union[Doc, Span]) -> float:
    """Calculate the mean word length of a document.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object

    Returns:
        float: mean word length
    """
    tokens_lengths = [len(t) for t in span]
    if tokens_lengths:
        return float(np.mean(tokens_lengths))
    return 0.0


def alpha_ratio(span: Union[Doc, Span]) -> float:
    """The percentage of spacy tokens in this document which contains at
    least one alphabetic character.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object

    Returns:
        float: alpha ratio
    """

    def contains_alpha(token):
        for char in token.text:
            if char.isalpha():
                return True
        return False

    token_contains_alpha = [contains_alpha(token) for token in span]
    if token_contains_alpha:
        return float(np.mean(token_contains_alpha))
    return 0.0


def proportion_bullet_points(  # pylint: disable=dangerous-default-value
    span: Union[Doc, Span],
    bullet_point: set = {"-", "*"},
) -> float:
    """Calculate the proportion of lines which start with a bullet points in a
    span.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object
        bullet_point (set): set of bullet points

    Returns:
        float: proportion of bullet points
    """
    # check if has extension _lines
    if not hasattr(span._, "lines"):
        lines = span.text.split("\n")
    else:
        lines = span._.lines
    line_starts_with_bullet = [
        line.strip().startswith(tuple(bullet_point)) for line in lines
    ]
    if line_starts_with_bullet:
        return float(np.mean(line_starts_with_bullet))
    return 0.0


def proportion_ellipsis(  # pylint: disable=dangerous-default-value
    span: Union[Doc, Span],
    ellipsis: set = {"…", "..."},
) -> float:
    """Calculate the proportion line which ends with an ellipsis in a span.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object
        ellipsis (set): set of ellipsis

    Returns:
        float: proportion of ellipsis
    """
    if not hasattr(span._, "lines"):
        lines = span.text.split("\n")
    else:
        lines = span._.lines

    line_ends_with_ellipsis = [line.strip().endswith(tuple(ellipsis)) for line in lines]
    if line_ends_with_ellipsis:
        return float(np.mean(line_ends_with_ellipsis))
    return 0.0


def get_ranges(arr: np.ndarray) -> List[Tuple[int, int]]:
    """Get ranges that evaluate to true a from boolean array, i.e.

    Example:
        >>> get_ranges(np.array([0, 1, 1, 0, 0, 1, 1]))
        [(1, 3), (5, 7)]
    """
    ranges = []
    start = None
    for i, val in enumerate(arr):
        if (val and start) is None:
            start = i
        elif not val and start is not None:
            ranges.append((start, i))
            start = None
    if start is not None:
        ranges.append((start, len(arr)))
    return ranges


def duplicate_paragraph_chr_fraction(span: Union[Doc, Span]) -> float:
    """Calculate the character fraction of duplicate paragraphs.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object

    Returns:
        float: The fraction of duplicate characters.
    """
    chr_len = len(span.text)
    if chr_len == 0:
        return 0.0

    if not hasattr(span._, "paragraphs"):
        paragraphs = span.text.split("\n\n")
    else:
        paragraphs = span._.paragraphs
    paragraph_counter = Counter(paragraphs)

    duplicate_chr = 0
    for t, c in paragraph_counter.items():
        if c > 1:
            duplicate_chr += len(t) * (c - 1)
    frac = duplicate_chr / chr_len
    return frac


def duplicate_line_chr_fraction(span: Union[Doc, Span]) -> float:
    """Calculate the character fraction of duplicate lines.

    Args:
        span (Union[Doc, Span]): A spaCy Doc or Span object

    Returns:
        float: The fraction of duplicate characters.
    """
    chr_len = len(span.text)
    if chr_len == 0:
        return 0.0

    if not hasattr(span._, "lines"):
        lines = span.text.split("\n")
    else:
        lines = span._.lines
    line_counter = Counter(lines)

    duplicate_chr = 0
    for t, c in line_counter.items():
        if c > 1:
            duplicate_chr += len(t) * (c - 1)
    frac = duplicate_chr / chr_len
    return frac


def symbol_to_word_ratio(span: Union[Span, Doc], symbol: str) -> float:
    """Calculate the ratio of symbols to words in a span.

    Args:
        span (Union[Span, Doc]): spaCy Span or Doc object
        ratio (float): ratio of symbols to words
        symbol (str): symbol to count

    Returns:
        float: ratio of symbols to words
    """
    n_symbol = span.text.count(symbol)
    n_words = sum(not (t.is_space or t.is_punct) for t in span)
    if n_words:
        return n_symbol / n_words
    return 0.0


def span_ngrams(
    span: Union[Span, Doc],
    ngram_range: Tuple[int, int],
) -> Dict[int, Dict[str, Union[int, List[Span]]]]:
    """Calculates the counts of n-grams in the specified range.

    Args:
        span (Union[Span, Doc]): A spaCy Span or Doc object.
        ngram_range (Tuple[int, int]): The n-gram range.

    Returns:
        Dict[int, Dict[str, Union[int, List[Span]]]]: A dictionary that for each n in
            the ngram range contains the counts of the n-grams as well as the spans of
            the n-grams.
    """
    max_len = len(span)
    lower, upper = ngram_range
    shingles_count = {  # type: ignore
        n: defaultdict(lambda: {"count": 0, "span": []})
        for n in range(lower, upper + 1)
    }
    for i, _ in enumerate(span):
        for ngram_size in range(lower, upper + 1):
            end = i + ngram_size
            if not end > max_len:
                ngram_span = span[i:end]
                ngram = ngram_span.text
                shingles_count[ngram_size][ngram]["count"] += 1  # type: ignore
                shingles_count[ngram_size][ngram]["span"].append(  # type: ignore
                    ngram_span,
                )
    return shingles_count  # type: ignore


def duplicate_ngram_fraction(
    span: Union[Span, Doc],
    ngram_range: Tuple[int, int],
) -> Dict[int, float]:
    """Calculates the character fraction of duplicate n-gram over the overall
    text, taking care not to count overlapping n-grams twice. This does not
    include spaces between the n-grams.

    Args:
        span (Union[Span, Doc]): A spaCy Span or Doc object.
        ngram_range (Tuple[int, int], optional): The n-gram range.

    Returns:
        Dict[int, float]: the fraction of duplicate characters for each
            n-gram size
    """
    max_len = len(span)
    chr_len = len(span.text)
    if chr_len == 0:
        return {n: 0.0 for n in range(ngram_range[0], ngram_range[1] + 1)}
    shingles_count = span_ngrams(span, ngram_range)
    duplicate_chr_fraction = {}
    for ngram_size, ngrams in shingles_count.items():
        # create a boolean array of the same length as the text
        # where True indicates that the token is a duplicate
        is_duplicate = np.zeros(max_len, dtype=bool)
        # set duplicate tokens to True
        for ngram, count in ngrams.items():
            if count["count"] > 1:  # type: ignore
                for ngram_span in count["span"]:  # type: ignore
                    is_duplicate[ngram_span.start : ngram_span.end] = True

        duplicate_chars = 0
        # get duplicate ranges from boolean array
        for start, end in get_ranges(is_duplicate):
            _span = span[start:end]
            duplicate_chars += _span.end_char - _span.start_char
        duplicate_chr_fraction[ngram_size] = duplicate_chars / chr_len
    return duplicate_chr_fraction


def top_ngram_chr_fraction(
    span: Union[Doc, Span],
    ngram_range: Tuple[int, int],
    min_count: int = 0,
) -> Dict[int, float]:
    """Calculates the character fraction of the top ngrams.

    Args:
        span (Union[Span, Doc]): A spaCy Span or Doc object.
        ngram_range (Tuple[int, int], optional): Range of n grams to examine.
        min_count (int): Minimum count of n-grams to before an n-gram is considered
            a top n-gram. Defaults to 0.

    Returns:
        Dict[int, float]: the fraction of duplicate characters for each
            n-gram size
    """
    # check if span has enough tokens within the range

    chr_len = len(span.text)
    if chr_len == 0:
        return {n: 0.0 for n in range(ngram_range[0], ngram_range[1] + 1)}

    ngram_counter = span_ngrams(span, ngram_range=ngram_range)
    top_ngram_chr_frac = {}
    for n in ngram_counter:
        # find the top n-gram
        if ngram_counter[n]:
            ngram, count_span = max(
                ngram_counter[n].items(),
                key=lambda x: x[1]["count"],  # type: ignore
            )
            count = count_span["count"]  # type: ignore
            if count >= min_count:
                # calculate the fraction of the top n-gram
                top_ngram_chr_frac[n] = (len(ngram) * count) / chr_len
            else:
                top_ngram_chr_frac[n] = 0.0
        else:
            top_ngram_chr_frac[n] = 0.0

    return top_ngram_chr_frac


def contains_string(span: Union[Span, Doc], string: str) -> bool:
    """Check if a span contains a string.

    Args:
        span (Union[Span, Doc]): A spaCy Span or Doc object.
        string (str): string to check for

    Returns:
        bool: True if span contains string
    """
    return string in span.text


def oov_ratio(span: Union[Span, Doc], vocab: Optional[Mapping] = None) -> float:
    """Calculates the out-of-vocabulary ratio.

    Args:
        span (Union[Span, Doc]): A spaCy Span or Doc object.
        vocab (Optional[Mapping], optional): A vocabulary to check against.
            If None, will use the spaCy vocab. Note that the spaCy vocab
            is not defined for small models. Defaults to None.

    Returns:
        float: the out-of-vocabulary ratio
    """
    len_span = len(span)
    if len_span == 0:
        return 0.0
    if vocab is None:
        return len([token for token in span if token.is_oov]) / len_span
    return len([token for token in span if token.text not in vocab]) / len_span


class Quality:
    """spaCy component for adding text quality metrics to the `Doc` and `Span`
    objects.

    Extracts metrics and returns them as a dictionary as the ._.quality
    attribute.
    """

    def __init__(  # pylint: disable=dangerous-default-value
        self,
        nlp: Language,
        name: str,
        top_ngram_range: Tuple[int, int],
        top_ngram_min_count: int,
        duplicate_n_gram_fraction_range: Tuple[int, int],
        vocab: Optional[Mapping],
        quality_thresholds: Optional[QualityThresholds] = None,
        force: bool = False,
    ):  # noqa: D107
        """Initialise components."""
        self.name = name
        self.force = force
        self.top_ngram_range = top_ngram_range
        self.top_ngram_min_count = top_ngram_min_count
        self.duplicate_n_gram_fraction_range = duplicate_n_gram_fraction_range

        if quality_thresholds is None:
            quality_thresholds = QualityThresholds()
        self.set_quality_thresholds(quality_thresholds)

        self.vocab = vocab

        self.set_extensions()

    def quality_setter(
        self,
        span: Union[Span, Doc],
    ) -> QualityOutput:
        """Apply quality functions to doc.

        Args:
            span (Union[Span, Doc]): spaCy span or doc object

        Returns:
            QualityOutput: The quality metrics
        """
        threshold = self.quality_thresholds

        thresholds_outputs: Dict[
            str,
            Union[Dict[str, ThresholdsOutput], ThresholdsOutput],
        ] = {}
        # filter with only one threshold
        getters = {
            # heuristic quality filters
            "n_stop_words": n_stop_words,
            "alpha_ratio": alpha_ratio,
            "mean_word_length": mean_word_length,
            "doc_length": len,
            "proportion_ellipsis": proportion_ellipsis,
            "proportion_bullet_points": proportion_bullet_points,
            # text repetition
            "duplicate_line_chr_fraction": duplicate_line_chr_fraction,
            "duplicate_paragraph_chr_fraction": duplicate_paragraph_chr_fraction,
        }

        for name, getter in getters.items():
            thresholds_outputs[name] = ThresholdsOutput(
                value=getter(span),  # type: ignore
                threshold=getattr(threshold, name),
            )

        thresholds_outputs["contains"] = {
            string: ThresholdsOutput(
                value=contains_string(span, string),
                threshold=threshold.contains.get(string, None),
            )
            for string in self.contains
        }
        thresholds_outputs["symbol_to_word_ratio"] = {
            symbol: ThresholdsOutput(
                value=symbol_to_word_ratio(span, symbol),
                threshold=threshold.symbol_to_word_ratio.get(symbol, None),
            )
            for symbol in self.symbols
        }

        chr_frac = top_ngram_chr_fraction(
            span,
            ngram_range=self.top_ngram_range,
            min_count=self.top_ngram_min_count,
        )

        thresholds_outputs["top_ngram_chr_fraction"] = {
            str(n_gram): ThresholdsOutput(
                value=frac,
                threshold=threshold.top_ngram_chr_fraction.get(
                    str(n_gram),
                    (None, None),
                ),
            )
            for n_gram, frac in chr_frac.items()
        }

        duplicate_ngram_chr_fraction = duplicate_ngram_fraction(
            span,
            ngram_range=self.duplicate_n_gram_fraction_range,
        )
        thresholds_outputs["duplicate_ngram_chr_fraction"] = {
            str(n_gram): ThresholdsOutput(
                value=frac,
                threshold=threshold.duplicate_ngram_chr_fraction.get(
                    str(n_gram),
                    (None, None),
                ),
            )
            for n_gram, frac in duplicate_ngram_chr_fraction.items()
        }

        # add oov_ratio if spacy model is not small or has a vocab
        # vector length is 0 for small models
        if span.vocab.vectors_length > 0 or self.vocab:
            value_oov = oov_ratio(span, self.vocab)
            thresholds_oov = threshold.oov_ratio
        else:
            value_oov = None
            thresholds_oov = (None, None)

        thresholds_outputs["oov_ratio"] = ThresholdsOutput(
            value=value_oov,
            threshold=thresholds_oov,
        )

        return QualityOutput(**thresholds_outputs)

    def quality_getter(self, span: Union[Span, Doc]) -> QualityOutput:
        """Get quality metrics from doc.

        Args:
            span (Union[Span, Doc]): spaCy span or doc object

        Returns:
            QualityOutput: The quality metrics
        """
        if not hasattr(span._, "_quality"):
            return self.quality_setter(span)
        return QualityOutput(**span._._quality)

    def set_quality(self, doc: Doc) -> None:
        """Set the quality attribute on a doc.

        Args:
            doc (Doc): spaCy doc object
        """
        # to allow the variable to json serializable we convert it to json
        # it is then converted back into a quality output object in the getter

        doc._._quality = self.quality_setter(doc).model_dump()
        doc._.passed_quality_check = self.passed_quality_thresholds(doc)

    def passed_quality_thresholds(self, span: Union[Span, Doc]) -> bool:
        """Check if a span passes the quality thresholds.

        Args:
            span (Union[Span, Doc]): spaCy span or doc object

        Returns:
            bool: True if span passes quality thresholds
        """
        quality_output = self.quality_getter(span)
        return quality_output.passed

    def set_extensions(self):
        """Set required extensions."""
        ext_name = "passed_quality_check"
        if not Span.has_extension(ext_name) or self.force is True:
            Span.set_extension(
                ext_name,
                getter=self.passed_quality_thresholds,
                force=True,
            )
        if not Doc.has_extension(ext_name) or self.force is True:
            Doc.set_extension(
                ext_name,
                getter=self.passed_quality_thresholds,
                force=True,
            )

        ext_name = "quality"
        if not Doc.has_extension(ext_name) or self.force is True:
            Doc.set_extension(ext_name, getter=self.quality_getter, force=True)
            Doc.set_extension("_" + ext_name, default=None, force=True)
        if not Span.has_extension(ext_name) or self.force is True:
            Span.set_extension(ext_name, getter=self.quality_getter, force=True)
            Span.set_extension("_" + ext_name, default=None, force=True)

    def set_quality_thresholds(self, thresholds: QualityThresholds) -> None:
        """Sets the quality thresholds.

        Args:
            thresholds (QualityThresholds): The desired quality thresholds.
        """
        self.quality_thresholds = thresholds
        self.contains = list(self.quality_thresholds.contains.keys())
        self.symbols = list(self.quality_thresholds.symbol_to_word_ratio.keys())

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        self.set_quality(doc)
        return doc


@Language.factory(
    "textdescriptives/quality",
    assigns=[
        "doc._.quality",
        "doc._.passed_quality_check",
        "span._.quality",
        "span._.passed_quality_check",
    ],
    default_config={
        "top_ngram_range": [2, 4],
        "top_ngram_min_count": 3,
        "duplicate_n_gram_fraction_range": [5, 10],
        "vocab": None,
        "force": True,
    },
)
def create_quality_component(
    nlp: Language,
    name: str,
    top_ngram_range: Tuple[int, int],
    top_ngram_min_count: int,
    duplicate_n_gram_fraction_range: Tuple[int, int],
    vocab: Optional[Mapping],
    force: bool = True,
) -> Callable[[Doc], Doc]:
    """Allows Quality to be added to a spaCy pipe using
    nlp.add_pipe("textdescriptives/quality").

    Adding this component to a pipeline sets the following attributes:

    - {Span/Doc}._.quality
    - {Span/Doc}._.passed_quality_check

    It also sets:

    - {Span/Doc}._.lines
    - {Span/Doc}._.paragraphs

    These are used to calculate some of the quality metrics. They can be overwritten if
    you e.g. wish lines to be split on "\\r\\n" instead of "\\n".

    A large part of the quality metrics were proposed by [1] and [2] for filtering
    out low quality text from large text corpora.

    References:
    - [1] Rae, J. W., Borgeaud, S., Cai, T., Millican, K., Hoffmann, J., Song, F., ... &
    Irving, G. (2021). Scaling language models: Methods, analysis & insights from
    training gopher. arXiv preprint arXiv:2112.11446.
    - [2] Raffel, C., Shazeer, N., Roberts, A., Lee, K., Narang, S., Matena, M., ... &
    Liu, P. J. (2020). Exploring the limits of transfer learning with a unified
    text-to-text transformer. J. Mach. Learn. Res., 21(140), 1-67.

    Args:
        nlp (Language): spaCy language object, does not need to be specified in the
            nlp.add_pipe call.
        name (str): name of the component. Can be optionally specified in the
            nlp.add_pipe call, using the name argument.
        top_ngram_range (Tuple[int]): range of n-grams to calculate the
            proportion of the top n-gram. Defaults to [2, 4].
        top_ngram_min_count (int): minimum number of times a n-gram must occur to
            be considered a top n-gram. Defaults to 3.
        duplicate_n_gram_fraction_range (Tuple[int]): range of n-grams to
            calculate the proportion of duplicate n-grams. Defaults to [5, 10].
        vocab (Optional[Mapping]): vocabulary to use for calculating the
            out-of-vocabulary ratio (`oov_ratio`). If None, will use the vocabulary
            of the spaCy model. Note, that small spaCy models do not have a
            vocabulary. The attribute will only be set if the vocabulary is not
            None or the spaCy model is medium or large.
        force (bool): whether to overwrite existing extensions. Defaults to True.


    Returns:
        Callable[[Doc], Doc]: the Quality component

    Example:
        >>> import spacy
        >>> from spacy_quality import Quality
        >>> nlp = spacy.blank("en")
        >>> nlp.add_pipe("quality")
        >>> doc = nlp("This is a test")
        >>> # extract quality metrics
        >>> doc._.quality
        >>> # check whether the document passed the quality thresholds
        >>> doc._.passed_quality_check
    """
    return Quality(
        nlp,
        name=name,
        top_ngram_range=top_ngram_range,
        top_ngram_min_count=top_ngram_min_count,
        duplicate_n_gram_fraction_range=duplicate_n_gram_fraction_range,
        quality_thresholds=None,
        vocab=vocab,
        force=force,
    )



""" Adds all components to a spaCy pipeline."""

from spacy.language import Language
from spacy.tokens import Doc


class TextDescriptives:
    """Utility spaCy v3.0 component to add all functionality from the
    TextDescriptives package to a `Doc` objects. See `DescriptiveStatistics`,
    `Readability`, and `DependencyDistance` for more details.

    Example:
        >>> import spacy
        >>> from textdescriptives import TextDescriptives
        >>> nlp = spacy.load("en_core_web_sm")
        >>> nlp.add_pipe("textdescriptives")
        >>> text = "This is a the first sentence. Here's a second, slightly longer one."
        >>> doc = nlp(text)
        >>> doc._.dependency_distance
        >>> doc._.readability
        >>> doc[0:4]._.token_length
    """

    def __init__(self, nlp: Language):
        """Don't do anything, just return the Doc."""

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        return doc


@Language.factory("textdescriptives/all")
def create_textdescriptives_component(nlp: Language, name: str):
    components = [
        k
        for k in Language.factories.keys()
        if k.startswith("textdescriptives") and k != "textdescriptives/all"
    ]

    for component in components:
        nlp.add_pipe(component, last=True)
    return TextDescriptives(nlp)
